""" webfilter/Rewriter.py

    Copyright (C) 2000  Bastian Kleineidam

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

We use the SGMLParser to parse the HTML stream in tokens.
For each token we look for applying rules.
"""
import sys,re,string
import webfilter,WebCleanerConf,htmlentitydefs
from xml.parsers.sgmllib import SlowSGMLParser
# the slow SGMLParser is much more stable than the fast one
SGMLParser = SlowSGMLParser
orders = [webfilter.FILTER_RESPONSE_MODIFY]
from webfilter import debug, error
from webfilter.get_dom import *
from webfilter.Rules import GetRuleFromNode, STARTTAG, ENDTAG, DATA, COMMENT

class Rewriter(webfilter.Filter):
    """This class can rewrite HTML tags.
    Add rewrite rules conveniently with the addMatcher function.
    Note: output is buffered to provide seamless filtering.
    """
    mimelist = ('text/html',)

    def __init__(self):
        self.rules = []
        self.comments = 1

    def configure(self, dom):
        nodes = dom.getElementsByTagName("rewrite")
        for node in nodes:
            rule = GetRuleFromNode(node)
            if rule.disable:
                debug("disable rewriter '%s'" % rule.title, 2)
            else:
                debug("enable rewriter '%s'" % rule.title, 2)
                self.rules.append(rule)
        nodes = dom.getElementsByTagName('nocomments')
        for node in nodes:
            if int(get_node_attr(node, 'disable')):
                continue
            self.comments = 0
            break

    def filter(self, data, **attrs):
        p = attrs['filter']
        p.feed(data)
        return p.flush()

    def finish(self, data, **attrs):
        p = attrs['filter']
        p.feed(data)
        return p.close()

    def getAttrs(self, headers):
        """We need a separate filter instance for stateful filtering"""
        p = HtmlFilter(self.rules, self.comments, webfilter._DEBUG_LEVEL)
        return {'filter': p}



class HtmlFilter(SGMLParser):
    def __init__(self, rules, comments, verbose=0):
        SGMLParser.__init__(self, verbose)
        self.rules = rules
        self.comments = comments
        self.data = ""
        self.rulestack = []
        self.buffer = []
        self.entitydefs = htmlentitydefs.entitydefs


    def buffer_append_data(self, data):
        """we have to make sure that we have no two following
        DATA things in the buffer. Why? well, just to be 100% sure
        that an ENCLOSED match really matches enclosed data.
        """
        if self.buffer and data[0]==DATA and self.buffer[-1][0]==DATA:
            self.buffer[-1][1] = self.buffer[-1][1] + data[1]
        else:
            self.buffer.append(data)


    def handle_cdata(self, d):
        """sgmlop.SGMLParser compatibility handler for cdata (?)"""
        debug("cdata: "+d, 3)

    def handle_charref(self, d):
        """sgmlop.SGMLParser compatibility handler for charref (?)"""
        debug("charref: "+d, 3)

    def handle_special(self, d):
        """sgmlop.SGMLParser compatibility handler for special (?)"""
        debug("special: "+d, 3)

    def handle_proc(self, d):
        """sgmlop.SGMLParser compatibility handler for proc (?)"""
        debug("proc: "+d, 3)

    def handle_data(self, d):
        """handler for data"""
        debug("data: "+d, 3)
        self.buffer_append_data([DATA, d])


    def flush(self):
        data = self.data
        if data:
            self.data = ""
        return data


    def buffer2data(self):
        """Append all tags of the buffer to the data"""
        for n in self.buffer:
            if n[0]==DATA:
                self.data = self.data + n[1]
            elif n[0]==COMMENT:
                self.data = self.data+"<!--"+n[1]+"-->"
            elif n[0]==STARTTAG:
                s = "<"+n[1]
                for name,val in n[2]:
                    s = s+' '+name+'="'+val+'"'
                self.data = self.data +s+">"
            elif n[0]==ENDTAG:
                self.data = self.data+"</"+n[1]+">"
            else:
                error("unknown buffer element %s" % n[0])
        self.buffer = []


    def handle_comment(self, data):
        """a comment. either delete it or print it, nothing more
	   because we dont filter inside comments"""
        debug("comment: "+data, 3)
        if self.comments:
            self.buffer.append((COMMENT,data))


    def unknown_starttag(self, tag, attrs):
        """We get a new start tag. New rules could be appended to the
        pending rules. No rules can be removed from the list."""
        debug("start tag: <" + tag + ">", 3)
        tobuffer = (STARTTAG,tag,attrs)
        rulelist = []
        for rule in self.rules:
            if rule.match_tag(tag) and rule.match_attrs(attrs):
                debug("matched rule %s" % rule.title, 3)
                if rule.start_sufficient:
                    tobuffer = rule.filter_tag(tag, attrs)
                    break
                else:
                    rulelist.append(rule)
        if rulelist:
            self.rulestack.append((len(self.buffer), rulelist))
        self.buffer_append_data(tobuffer)
        if not self.rulestack:
            self.buffer2data()


    def unknown_endtag(self, tag):
        """We know the following: if a rule matches, it must be
        the one on the top of the stack. So we look only at the top
        rule.
	If it matches and the rule stack is now empty we can flush
	the buffer (by calling buffer2data)"""
        debug('end tag: </' + tag + '>', 3)
        self.buffer.append((ENDTAG, tag))
        if self.rulestack and self.rulestack[-1][1][0].match_tag(tag):
            i, rulelist = self.rulestack.pop()
            for rule in rulelist:
                if rule.match_complete(i, self.buffer):
                    rule.filter_complete(i, self.buffer)
                    break
        if not self.rulestack:
            self.buffer2data()


    def unknown_entityref(self, ref):
        """this occurs usually with unquoted ampersands. never mind,
           we just print it out."""
        debug('*** unknown entity ref: &' + ref + ';\n')
        self.buffer_append_data([DATA, "&"+ref])


    def unknown_charref(self, ref):
        """same as unknown_entity. print it out"""
        debug('*** unknown char ref: &#' + ref + ';\n')
        self.buffer_append_data([DATA, "&"+ref])


    def close(self):
        SGMLParser.close(self)
        self.buffer2data()
        return self.flush()
